import os
import json
import msal
import requests
import uuid
import logging
from pprint import pformat

from flask import Flask, render_template, abort, jsonify, session, url_for, request, redirect
from flask_session import Session  # https://pythonhosted.org/Flask-Session
from flask_cors import CORS

from p2.ui import settings
from p2.ui.core import Core
from p2.tools.packages import get_p2_packages

app = Flask(__name__)
CORS(app)
app.logger.setLevel(logging.DEBUG)
core = Core()

app.config.from_object(settings)
Session(app)

# This section is needed for url_for("foo", _external=True) to automatically
# generate http scheme when this sample is running on localhost,
# and to generate https scheme when it is deployed behind reversed proxy.
# See also https://flask.palletsprojects.com/en/1.0.x/deploying/wsgi-standalone/#proxy-setups
from werkzeug.middleware.proxy_fix import ProxyFix
app.wsgi_app = ProxyFix(app.wsgi_app, x_proto=1, x_host=1)

tenant_id = os.getenv("TENANT_ID")
if tenant_id is None:
    raise Exception("No TENANT_ID defined")
authority = f"{settings.AUTHORITY}/{tenant_id}"

client_id = os.getenv("CLIENT_ID")
if client_id is None:
    raise Exception("No CLIENT_ID defined")

@app.route('/packages')
def packages():
    return jsonify(get_p2_packages())

@app.route('/api/js')
def js():
    return jsonify(core.get_js())

@app.route('/api/css')
def css():
    return jsonify(core.get_css())

@app.route('/')
def index():
    session["state"] = str(uuid.uuid4())
    auth_url = _build_auth_url(
        scopes    = settings.DELEGATED_PERMISSIONS,
        state     = session["state"])

    app.logger.debug('[/]: auth_url: %s', auth_url)
    data = {
        'js_files': core.get_js(),
        'css_files': core.get_css(),
        'menubar': core.get_menubar() or [],
        'user': session.get('user') or os.getenv('P2_UI_NO_AUTH'),
        'auth_url': auth_url
    }
    return render_template('main.html', **data)

@app.route("/logout")
def logout():
    session.clear()  # Wipe out user and its token cache from session
    redirect_url = url_for("index", _external=True)
    app.logger.debug("[/logout]: Redirect url: %s", redirect_url)
    return redirect(authority + "/oauth2/v2.0/logout" + "?post_logout_redirect_uri=" + redirect_url)

@app.route(settings.REDIRECT_PATH)
def authorized():
    if request.args.get('code'):
        cache = _load_cache()
        redirect_url=url_for("authorized", _external=True)
        app.logger.debug("[/getToken]: Redirect url: %s", redirect_url)
        result = _build_msal_app(cache=cache).acquire_token_by_authorization_code(
            request.args['code'],
            scopes=settings.DELEGATED_PERMISSIONS,
            redirect_uri=redirect_url
        )
        app.logger.debug('[/getToken]: result: %s', pformat(result))
        if "error" in result:
            return render_template("auth_error.html", result=result)
        session["user"] = result.get("id_token_claims")
        _save_cache(cache)

    data = {
        'js_files': core.get_js(),
        'css_files': core.get_css(),
        'menubar': core.get_menubar(),
        'user': session['user']
    }
    return render_template('main.html', **data)

@app.route("/select/<table>", methods=['POST'])
def select(table):
    """ Performs a request action against a p2.store table and returns
    back the result.
    The table is passed as query parameter.
    Form parameters:
        - where: the where clause as expected by postgres or p2.store clients
        - sort: the sort clause
        - limit: number of records to send back

    Returns 200 and the JSON-encoded list of records or 500 in case of errors
    """
    if table is None:
        raise Exception("table not set")
    args = dict(request.args)
    data = dict(request.form) or json.loads(request.data)
    params = json.loads(data['params'])
    app.logger.debug(f"Selecting records in table {table}")
    app.logger.debug(f"Form parameters: {params}")

    try:
        records = core.client.select(table=table, **params)
        if records is None:
            raise exception("Error selecting from table")
        return jsonify(records)
    except Exception as e:
        abort(500, e)

@app.route("/update/<table>", methods=['PUT'])
def update(table):
    """ Updates the table passed in the path """
    if table is None:
        raise Exception("table not set")
    args = dict(request.args)
    data = dict(request.form) or json.loads(request.data)
    params = json.loads(data['params'])
    app.logger.debug(f"Updating  table {table}")
    app.logger.debug(f"Form parameters: {params}")

    try:
        where  = params['where']
        fields = params['fields']
        resp = core.client.update(table, fields, where)
        if resp is None:
            raise exception(f"Error updating table {table}")
        return jsonify(resp)
    except Exception as e:
        abort(500, e)

@app.route("/update_on_conflict/<table>", methods=['POST'])
def update_on_conflict(table):
    """ Tries to insert a record in a table, if Updates the table passed in the path """
    if table is None:
        raise Exception("table not set")
    args = dict(request.args)
    data = dict(request.form) or json.loads(request.data)
    params = json.loads(data['params'])
    app.logger.debug(f"Inserting or updating table {table}")
    app.logger.debug(f"Form parameters: {params}")

    try:
        conflicting_cols  = params['conflicting_cols']
        record = params['record']
        resp = core.client.update_on_conflict(
            table = table,
            conflicting_cols = conflicting_cols,
            record = record)
        if resp is None:
            raise exception(f"Error updating table {table}")
        return jsonify(resp)
    except Exception as e:
        abort(500, e)


#----- utility functions
def _load_cache():
    cache = msal.SerializableTokenCache()
    if session.get("token_cache"):
        cache.deserialize(session["token_cache"])
    return cache

def _save_cache(cache):
    if cache.has_state_changed:
        session["token_cache"] = cache.serialize()

def _build_msal_app(cache=None):
    return msal.ConfidentialClientApplication(
        client_id=client_id,
        authority=authority,
        client_credential=os.getenv('CLIENT_SECRET', ''),
        token_cache=cache,
        verify=False
    )

def _build_auth_url(scopes=None, state=None):
    redirect_url=url_for("authorized", _external=True)
    app.logger.debug("[_build_auth_url]: Redirect url: %s", redirect_url)
    return _build_msal_app().get_authorization_request_url(
        scopes or [],
        state=state or str(uuid.uuid4()),
        redirect_uri=redirect_url
    )


if __name__ == '__main__':
    app.run()
